import * as babel from '@babel/core';
import { buildExternalHelpers } from '@babel/core';
import getFixtures, { resolveOptionPluginOrPreset } from '@babel/helper-fixtures';
import { codeFrameColumns } from '@babel/code-frame';
import { eachMapping, TraceMap } from '@jridgewell/trace-mapping';
import assert from 'assert';
import fs from 'fs';
import path from 'path';
import vm from 'vm';
import LruCache from 'lru-cache';
import { fileURLToPath } from 'url';
import { diff } from 'jest-diff';
import { createRequire } from 'module';
import checkDuplicateNodes from '@babel/helper-check-duplicate-nodes';

function assertNoOwnProperties(obj) {
  expect(Object.getOwnPropertyNames(obj)).toHaveLength(0);
}
function multiline(arr) {
  return arr.join("\n");
}

var helpers = /*#__PURE__*/Object.freeze({
  __proto__: null,
  assertNoOwnProperties: assertNoOwnProperties,
  multiline: multiline
});

const CONTEXT_SIZE = 4;
const LOC_SIZE = 10;
const CONTENT_SIZE = 15;
const padStart = "".padStart ? Function.call.bind("".padStart) : (str, len, pad) => pad.repeat(Math.max(0, len - str.length)) + str;
const padEnd = "".padStart ? Function.call.bind("".padEnd) : (str, len, pad) => str + pad.repeat(Math.max(0, len - str.length));
function simpleCodeFrameRange(lines, line, colStart, colEnd) {
  colEnd = Math.min(colEnd, lines[line - 1].length);
  const start = Math.max(colStart - CONTEXT_SIZE, 0);
  const end = Math.min(colEnd + CONTEXT_SIZE, lines[line - 1].length);
  const markerSize = colEnd - colStart;
  const marker = markerSize === 0 ? "><" : " " + "^".repeat(markerSize);
  const markerPadding = colStart - start - 1;
  const code = lines[line - 1].slice(start, end);
  const loc = padStart(`(${line}:${colStart}-${colEnd}) `, LOC_SIZE, " ");
  return loc + code + "\n" + " ".repeat(markerPadding + loc.length) + marker;
}
function joinMultiline(left, right, leftLen) {
  const leftLines = left.split("\n");
  const rightLines = right.split("\n");
  leftLen ??= leftLines.reduce((len, line) => Math.max(len, line.length), 0);
  const linesCount = Math.max(leftLines.length, rightLines.length);
  let res = "";
  for (let i = 0; i < linesCount; i++) {
    if (res !== "") res += "\n";
    if (i < leftLines.length) res += padEnd(leftLines[i], leftLen, " ");else res += " ".repeat(leftLen);
    if (i < rightLines.length) res += rightLines[i];
  }
  return res;
}
function visualize(output, map) {
  const sourcesLines = new Map(map.sources.map((source, index) => [source, map.sourcesContent[index].split("\n")]));
  const outputLines = output.split("\n");
  const ranges = [];
  let prev = null;
  eachMapping(new TraceMap(map), mapping => {
    if (prev === null) {
      prev = mapping;
      return;
    }
    const original = {
      from: {
        line: prev.originalLine,
        column: prev.originalColumn
      },
      to: {
        line: mapping.originalLine,
        column: mapping.originalColumn
      }
    };
    const generated = {
      from: {
        line: prev.generatedLine,
        column: prev.generatedColumn
      },
      to: {
        line: mapping.generatedLine,
        column: mapping.generatedColumn
      }
    };
    if (original.from.line !== original.to.line) {
      original.to.line = original.from.line;
      original.to.column = Infinity;
    } else if (original.to.column < original.from.column) {
      original.to.column = original.from.column;
    }
    if (generated.from.line !== generated.to.line) {
      generated.to.line = generated.from.line;
      generated.to.column = Infinity;
    } else if (generated.to.column < generated.from.column) {
      generated.to.column = generated.from.column;
    }
    ranges.push({
      original,
      generated,
      source: prev.source
    });
    prev = mapping;
  });
  if (prev.originalLine) {
    ranges.push({
      original: {
        from: {
          line: prev.originalLine,
          column: prev.originalColumn
        },
        to: {
          line: prev.originalLine,
          column: Infinity
        }
      },
      generated: {
        from: {
          line: prev.generatedLine,
          column: prev.generatedColumn
        },
        to: {
          line: prev.generatedLine,
          column: Infinity
        }
      },
      source: prev.source
    });
  }
  for (let i = ranges.length - 1; i >= 0; i--) {
    const {
      original
    } = ranges[i];
    if (original.from.column === original.to.column && original.to.column < ranges[i + 1].original.to.column) {
      original.to.column = ranges[i + 1].original.to.column;
    }
  }
  const res = ranges.map(({
    original,
    generated,
    source
  }) => {
    const input = simpleCodeFrameRange(sourcesLines.get(source), original.from.line, original.from.column, original.to.column);
    const output = simpleCodeFrameRange(outputLines, generated.from.line, generated.from.column, generated.to.column);
    return joinMultiline(joinMultiline(input, " <--  ", LOC_SIZE + CONTEXT_SIZE * 2 + CONTENT_SIZE), output);
  });
  return res.join("\n\n");
}

const require = createRequire(import.meta.url);
const EXTERNAL_HELPERS_VERSION = "7.100.0";
const cachedScripts = new LruCache({
  max: 10
});
const contextModuleCache = new WeakMap();
const sharedTestContext = createContext();
function transformWithoutConfigFile(code, opts) {
  return babel.transformSync(code, Object.assign({
    configFile: false,
    babelrc: false
  }, opts));
}
function transformAsyncWithoutConfigFile(code, opts) {
  return babel.transformAsync(code, Object.assign({
    configFile: false,
    babelrc: false
  }, opts));
}
function createContext() {
  const context = vm.createContext(Object.assign({}, helpers, {
    process: process,
    transform: transformWithoutConfigFile,
    transformAsync: transformAsyncWithoutConfigFile,
    setTimeout: setTimeout,
    setImmediate: setImmediate,
    expect
  }));
  context.global = context;
  const moduleCache = Object.create(null);
  contextModuleCache.set(context, moduleCache);
  runCacheableScriptInTestContext(path.join(path.dirname(fileURLToPath(import.meta.url)), "babel-helpers-in-memory.js"), buildExternalHelpers, context, moduleCache);
  return context;
}
function runCacheableScriptInTestContext(filename, srcFn, context, moduleCache) {
  let cached = cachedScripts.get(filename);
  if (!cached) {
    const code = `(function (exports, require, module, __filename, __dirname) {\n${srcFn()}\n});`;
    cached = {
      code,
      cachedData: undefined
    };
    cachedScripts.set(filename, cached);
  }
  let script;
  {
    script = new vm.Script(cached.code, {
      filename,
      lineOffset: -1,
      cachedData: cached.cachedData
    });
    cached.cachedData = script.createCachedData();
  }
  const module = {
    id: filename,
    exports: {}
  };
  moduleCache[filename] = module;
  const req = id => runModuleInTestContext(id, filename, context, moduleCache);
  const dirname = path.dirname(filename);
  script.runInContext(context).call(module.exports, module.exports, req, module, filename, dirname);
  return module;
}
function runModuleInTestContext(id, relativeFilename, context, moduleCache) {
  const filename = require.resolve(id, {
    paths: [path.dirname(relativeFilename)]
  });
  if (filename === id) return require(id);
  if (moduleCache[filename]) return moduleCache[filename].exports;
  return runCacheableScriptInTestContext(filename, () => fs.readFileSync(filename, "utf8"), context, moduleCache).exports;
}
function runCodeInTestContext(code, opts, context = sharedTestContext) {
  const filename = opts.filename;
  const dirname = path.dirname(filename);
  const moduleCache = contextModuleCache.get(context);
  const req = id => runModuleInTestContext(id, filename, context, moduleCache);
  const module = {
    id: filename,
    exports: {}
  };
  const oldCwd = process.cwd();
  try {
    if (opts.filename) process.chdir(path.dirname(opts.filename));
    const src = `(function(exports, require, module, __filename, __dirname, opts) {\n${code}\n});`;
    return vm.runInContext(src, context, {
      filename,
      displayErrors: true,
      lineOffset: -1
    })(module.exports, req, module, filename, dirname, opts);
  } finally {
    process.chdir(oldCwd);
  }
}
async function maybeMockConsole(validateLogs, run) {
  const actualLogs = {
    stdout: "",
    stderr: ""
  };
  if (!validateLogs) return {
    result: await run(),
    actualLogs
  };
  const spy1 = jest.spyOn(console, "log").mockImplementation(msg => {
    actualLogs.stdout += `${msg}\n`;
  });
  const spy2 = jest.spyOn(console, "warn").mockImplementation(msg => {
    actualLogs.stderr += `${msg}\n`;
  });
  try {
    return {
      result: await run(),
      actualLogs
    };
  } finally {
    spy1.mockRestore();
    spy2.mockRestore();
  }
}
async function run(task) {
  const {
    actual,
    expect: expected,
    exec,
    options: opts,
    doNotSetSourceType,
    optionsDir,
    validateLogs,
    ignoreOutput,
    stdout,
    stderr
  } = task;
  function getOpts(self) {
    const newOpts = Object.assign({
      ast: true,
      cwd: path.dirname(self.loc),
      filename: self.loc,
      filenameRelative: self.filename,
      sourceFileName: self.filename
    }, doNotSetSourceType ? {} : {
      sourceType: "script"
    }, {
      babelrc: false,
      configFile: false,
      inputSourceMap: task.inputSourceMap || undefined
    }, opts);
    return resolveOptionPluginOrPreset(newOpts, optionsDir);
  }
  let execCode = exec.code;
  let result;
  let resultExec;
  if (execCode) {
    const context = createContext();
    const execOpts = getOpts(exec);
    ({
      result
    } = await maybeMockConsole(validateLogs, () => babel.transformAsync(execCode, execOpts)));
    checkDuplicateNodes(result.ast);
    execCode = result.code;
    try {
      resultExec = runCodeInTestContext(execCode, execOpts, context);
    } catch (err) {
      err.message = `${exec.loc}: ${err.message}\n` + codeFrameColumns(execCode, {});
      throw err;
    }
  }
  const inputCode = actual.code;
  const expectedCode = expected.code;
  if (!execCode || inputCode) {
    let actualLogs;
    ({
      result,
      actualLogs
    } = await maybeMockConsole(validateLogs, () => babel.transformAsync(inputCode, getOpts(actual))));
    const outputCode = normalizeOutput(result.code);
    checkDuplicateNodes(result.ast);
    if (!ignoreOutput) {
      if (!expected.code && outputCode && !opts.throws && fs.statSync(path.dirname(expected.loc)).isDirectory() && !process.env.CI) {
        const expectedFile = expected.loc.replace(/\.m?js$/, result.sourceType === "module" ? ".mjs" : ".js");
        console.log(`New test file created: ${expectedFile}`);
        fs.writeFileSync(expectedFile, `${outputCode}\n`);
        if (expected.loc !== expectedFile) {
          try {
            fs.unlinkSync(expected.loc);
          } catch (e) {}
        }
      } else {
        validateFile(outputCode, expected.loc, expectedCode);
        if (inputCode) {
          expect(expected.loc).toMatch(result.sourceType === "module" ? /\.mjs$/ : /\.js$/);
        }
      }
    }
    if (validateLogs) {
      const normalizationOpts = {
        normalizePathSeparator: true,
        normalizePresetEnvDebug: task.taskDir.includes("babel-preset-env")
      };
      validateFile(normalizeOutput(actualLogs.stdout, normalizationOpts), stdout.loc, stdout.code.replace(/\n\s*syntax-(?!import-attributes|import-assertions).*/g, ""));
      validateFile(normalizeOutput(actualLogs.stderr, normalizationOpts), stderr.loc, stderr.code);
    }
  }
  if (task.validateSourceMapVisual === true) {
    const visual = visualize(result.code, result.map);
    try {
      expect(visual).toEqual(task.sourceMapVisual.code);
    } catch (e) {
      if (!process.env.OVERWRITE && task.sourceMapVisual.code) throw e;
      console.log(`Updated test file: ${task.sourceMapVisual.loc}`);
      fs.writeFileSync(task.sourceMapVisual.loc ?? task.taskDir + "/source-map-visual.txt", visual + "\n");
    }
  }
  if (opts.sourceMaps === true) {
    try {
      expect(result.map).toEqual(task.sourceMap);
    } catch (e) {
      if (!process.env.OVERWRITE && task.sourceMap) throw e;
      console.log(`Updated test file: ${task.sourceMapFile.loc}`);
      fs.writeFileSync(task.sourceMapFile.loc, JSON.stringify(result.map, null, 2));
    }
  }
  if (execCode && resultExec) {
    return resultExec;
  }
}
function validateFile(actualCode, expectedLoc, expectedCode) {
  if (actualCode !== expectedCode) {
    if (process.env.OVERWRITE) {
      console.log(`Updated test file: ${expectedLoc}`);
      fs.writeFileSync(expectedLoc, `${actualCode}\n`);
      return;
    }
    throw new Error(`Expected ${expectedLoc} to match transform output.\n` + `To autogenerate a passing version of this file, delete ` + ` the file and re-run the tests.\n\n` + `Diff:\n\n${diff(expectedCode, actualCode, {
      expand: false
    })}`);
  }
}
function escapeRegExp(string) {
  return string.replace(/[|\\{}()[\]^$+*?.]/g, "\\$&");
}
function normalizeOutput(code, {
  normalizePathSeparator = false,
  normalizePresetEnvDebug = false
} = {}) {
  const projectRoot = path.resolve(path.dirname(fileURLToPath(import.meta.url)), "../../../");
  const cwdSymbol = "<CWD>";
  let result = code.trim().replace(new RegExp(escapeRegExp(projectRoot), "g"), cwdSymbol);
  if (process.platform === "win32") {
    result = result.replace(new RegExp(escapeRegExp(projectRoot.replace(/\\/g, "/")), "g"), cwdSymbol).replace(new RegExp(escapeRegExp(projectRoot.replace(/\\/g, "\\\\")), "g"), cwdSymbol);
    if (normalizePathSeparator) {
      result = result.replace(/<CWD>[\w\\/.-]+/g, path => path.replace(/\\\\?/g, "/"));
    }
  }
  return result;
}
function index (fixturesLoc, name, suiteOpts = {}, taskOpts = {}, dynamicOpts) {
  const suites = getFixtures(fixturesLoc);
  for (const testSuite of suites) {
    if (suiteOpts.ignoreSuites?.includes(testSuite.title)) continue;
    describe(name + "/" + testSuite.title, function () {
      for (const task of testSuite.tests) {
        if (suiteOpts.ignoreTasks?.includes(task.title) || suiteOpts.ignoreTasks?.includes(testSuite.title + "/" + task.title)) {
          continue;
        }
        const testFn = task.disabled ? it.skip : it;
        testFn(task.title, async function () {
          const runTask = () => run(task);
          if ("sourceMap" in task.options) {
            throw new Error("`sourceMap` option is deprecated. Use `sourceMaps` instead.");
          }
          if ("sourceMaps" in task.options === false) {
            task.options.sourceMaps = !!task.sourceMap;
          }
          Object.assign(task.options, taskOpts);
          if (dynamicOpts) dynamicOpts(task.options, task);
          if (task.externalHelpers) {
            (task.options.plugins ??= []).push(["external-helpers", {
              helperVersion: EXTERNAL_HELPERS_VERSION
            }]);
          }
          const throwMsg = task.options.throws;
          if (throwMsg) {
            delete task.options.throws;
            await assert.rejects(runTask, function (err) {
              assert.ok(throwMsg === true || err.message.includes(throwMsg), `
Expected Error: ${throwMsg}
Actual Error: ${err.message}`);
              return true;
            });
          } else {
            return runTask();
          }
        });
      }
    });
  }
}

export { index as default, runCodeInTestContext };
//# sourceMappingURL=index.js.map
